import time
import os
import os
import json
import hashlib
from pathlib import Path
from typing import Union, List
from typing import List
from cachetools import LRUCache
from functools import lru_cache
from .cache_utils import get_cache_file

# Global cache dictionary using LRUCache
cache_dict = {}
# Adjust this value based on your memory constraints and requirements
MAX_CACHE_SIZE = 100000  # Example: 100k items

class LRUCacheManager:
    def __init__(self, maxsize=MAX_CACHE_SIZE):
        self.cache = LRUCache(maxsize=maxsize)

    def get(self, key):
        if key not in self.cache:
            return None
        return self.cache.get(key)

    def set(self, key, value):
        self.cache[key] = value

    def __getitem__(self, key):
        return self.get(key)
    
    def __setitem__(self, key, value):
        self.set(key, value)
    
    def __del__(self):
        self.save()

def load_cache(model_name, cache_dir=None):
    global cache_dict
    if model_name not in cache_dict:
        cache_dict[model_name] = LRUCacheManager()
        cache_file = get_cache_file(model_name, cache_dir)
        if cache_file.exists():
            print("Cache file exists at:", cache_file.absolute())
            print(f"Loading cache from {cache_file}")
            with open(cache_file, "r") as f:
                for line in f:
                    item = json.loads(line)
                    key = list(item.keys())[0]
                    value = list(item.values())[0]
                    # cache_dict[model_name][key] = {"output": value["output"]}
                    cache_dict[model_name][key] = value
    return cache_dict[model_name]




import json
import os
from pathlib import Path
from .cache_utils import get_cache_file
from collections import defaultdict
cache_dict = {}
def load_cache(model_name, cache_dir=None):
    global cache_dict
    if model_name not in cache_dict:
        cache_file = get_cache_file(model_name, cache_dir)
        if cache_file.exists():
            print("Cache file exists at:", cache_file.absolute())
            print(f"Loading cache from {cache_file}")
            with open(cache_file, "r") as f:
                model_cache_dict = [json.loads(line) for line in f.readlines()]
            model_cache_dict = {list(item.keys())[0]: list(item.values())[0] for item in model_cache_dict}
            # only keep the output in the value
            cache_dict[model_name] = defaultdict(lambda: None)
            for key, value in model_cache_dict.items():
                cache_dict[model_name][key] = value["output"]
        else:
            cache_dict[model_name] = {}
    return cache_dict[model_name]


import os
import time
from .utils import with_timeout
together_client = None
def call_worker_together(messages, model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    from together import Together
    global together_client
    if together_client is None:
        together_client = Together(api_key=os.environ.get("TOGETHER_API_KEY"), timeout=timeout)

    if model_name.startswith("together_"):
        model_name = model_name.replace("together_", "")
    
    new_messages = []
    if conv_system_msg:
        new_messages.append({"role": "system", "content": conv_system_msg})
    for i, message in enumerate(messages):
        new_messages.append({"role": "user" if i % 2 == 0 else "assistant", "content": message})
    
    stream = generate_kwargs.get("stream", False)
    if stream and "n" in generate_kwargs:
        generate_kwargs.pop("n")
        
    @with_timeout(timeout)
    def get_response():
        max_retry_for_unbound_local_error = 10
        retry_count = 0
        while True:
            try:
                completion = together_client.chat.completions.create(
                    model=model_name,
                    messages=new_messages,
                    **generate_kwargs,
                )
                break
            except UnboundLocalError as e:
                time.sleep(0.2)
                retry_count += 1
                if retry_count >= max_retry_for_unbound_local_error:
                    
                    raise e
                continue
        if not stream:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices]
            else:
                return completion.choices[0].message.content
        else:
            def generate_stream():
                for chunk in completion:
                    if chunk.choices[0].delta.content is not None:
                        yield chunk.choices[0].delta.content
            return generate_stream()
    return get_response()

def call_worker_together_completion(prompt:str, model_name, timeout:int=60, **generate_kwargs) -> str:
    from together import Together
    global together_client
    if together_client is None:
        together_client = Together(api_key=os.environ.get("TOGETHER_API_KEY"), timeout=timeout)

    if model_name.startswith("together_"):
        model_name = model_name.replace("together_", "")
    
    stream = generate_kwargs.get("stream", False)
    if stream and "n" in generate_kwargs:
        generate_kwargs.pop("n")
    @with_timeout(timeout)
    def get_response():
        completion = together_client.completions.create(
            model=model_name,
            prompt=prompt,
            **generate_kwargs,
        )
        if not stream:
            if len(completion.choices) > 1:
                return [c.text for c in completion.choices]
            else:
                return completion.choices[0].text
        else:
            def generate_stream():
                for chunk in completion:
                    if chunk.choices[0].delta.content is not None:
                        yield chunk.choices[0].delta.content
            return generate_stream()
    return get_response()


import os
import time
import torch
import random
import json
import openai
import vllm
import signal
import regex as re
from pathlib import Path
from typing import List
from .utils import SubprocessMonitor, ChatTokenizer, with_timeout, get_function_arg_names
from huggingface_hub import HfApi, hf_hub_download, snapshot_download
worker_initiated = False


chat_tokenizers = {}
def launch_vllm_worker(
    model_name: str,
    num_gpus: int=None,
    gpu_ids: List[int]=None,
    dtype: str="auto",
    quantization: str=None,
    port: int=34200,
    host: str="127.0.0.1",
    root_path: str=None,
    additional_args: List[str]=[],
) -> str:
    """
    Launch a model worker and return the address
    Args:
        model_name: the model name to launch
    Returns:
        the address of the launched model
    """
    print(f"Launching model {model_name}")
    worker_addr = f"http://{host}:{port}"
    log_file = Path(os.path.abspath(__file__)).parent / "logs" / f"{model_name}.log"
    log_file.parent.mkdir(parents=True, exist_ok=True)
    if gpu_ids:
        num_gpus = len(gpu_ids)
    else:
        if not num_gpus:
            num_gpus = torch.cuda.device_count()
            print(f"Warning: num_gpus or gpu_ids not provided, using {num_gpus} GPUs")
        gpu_ids = list(range(num_gpus))
    env = os.environ.copy()
    # Set the CUDA_VISIBLE_DEVICES environment variable
    env["CUDA_VISIBLE_DEVICES"] = ",".join([str(gpu_id) for gpu_id in gpu_ids])
    if "gemma-2" in model_name:
        env["VLLM_ATTENTION_BACKEND"] = "FLASHINFER"
    else:
        env["VLLM_ATTENTION_BACKEND"] = "FLASH_ATTN"
    print(num_gpus, gpu_ids)
    
    model_path = Path(model_name)
    if model_path.exists() and ((model_path / "config.json").exists() or (model_path / "adapter_config.json").exists()):
        if (model_path / "adapter_config.json").exists():
            print(f"Loading local model {model_name} with adapter")
            use_lora = True
            with open(model_path / "adapter_config.json") as f:
                adapter_config = json.load(f)
            adapter_path = model_path
            base_model_name_or_path = adapter_config["base_model_name_or_path"]
        elif (model_path / "config.json").exists():
            print(f"Loading local model {model_name}")
            use_lora = False
            adapter_path = None
            base_model_name_or_path = model_name
        else:
            raise ValueError(f"no config.json or adapter_config.json found in model {model_name}")
    else:
        # check whether there is a adapter_config.json
        api = HfApi()
        model_info = api.model_info(model_name)
        model_files = [x.rfilename for x in model_info.siblings]
        if "adapter_config.json" in model_files:
            use_lora = True
            adapter_path = snapshot_download(model_name)
            adapter_config_path = Path(adapter_path) / "adapter_config.json"
            with open(adapter_config_path) as f:
                adapter_config = json.load(f)
            base_model_name_or_path = adapter_config["base_model_name_or_path"]
            print(f"Loading model from Hugging Face {model_name} with adapter")
        elif "config.json" in model_files:
            use_lora = False
            adapter_path = None
            base_model_name_or_path = model_name
            print(f"Loading model from Hugging Face {model_name}")
        else:
            raise ValueError(f"no config.json or adapter_config.json found in model {model_name}")
        
    # python -m vllm.entrypoints.openai.api_server \
    # --model meta-llama/Llama-2-7b-hf \
    # --enable-lora \
    # --lora-modules sql-lora=~/.cache/huggingface/hub/models--yard1--llama-2-7b-sql-lora-test/
    if use_lora:
        lora_args = [
            "--enable-lora",
            "--lora-modules", f"{model_name}={adapter_path}",
            "--max-loras", "1",
            "--max-lora-rank", str(adapter_config["r"])
        ]
    else:
        lora_args = []
    if quantization:
        available_quantizations = "aqlm,awq,deepspeedfp,tpu_int8,fp8,fbgemm_fp8,marlin,gguf,gptq_marlin_24,gptq_marlin,awq_marlin,gptq,squeezellm,compressed-tensors,bitsandbytes,qqq,experts_int8"
        available_quantizations = available_quantizations.split(",")
        if quantization not in available_quantizations:
            raise ValueError(f"quantization {quantization} not in available quantizations: {available_quantizations}")
        lora_args += ["--quantization", quantization]
        if quantization == "bitsandbytes":
            lora_args += ["--load-format", "bitsandbytes", "--enforce-eager"]
    # python -m vllm.entrypoints.openai.api_server --model NousResearch/Meta-Llama-3-8B-Instruct --dtype auto --api-key token-abc123
    proc = SubprocessMonitor([
        "python3", "-m", "vllm.entrypoints.openai.api_server",
        "--model", base_model_name_or_path,
        "--dtype", dtype,
        # "--api-key", "vllm-engine-token",
        "--port", str(port),
        "--host", host,
        "--tensor-parallel-size", str(num_gpus),
        "--disable-log-requests",
        "--trust-remote-code",
    ] + (["--root-path", root_path] if root_path else [])
    + lora_args + additional_args, env=env)
    print(f"Launched VLLM model {model_name} at address {worker_addr} with CUDA_VISIBLE_DEVICES={env['CUDA_VISIBLE_DEVICES']}")
    if model_name not in chat_tokenizers:
        chat_tokenizers[model_name] = ChatTokenizer(base_model_name_or_path)
    if base_model_name_or_path not in chat_tokenizers:
        chat_tokenizers[base_model_name_or_path] = ChatTokenizer(base_model_name_or_path)
    return f"http://127.0.0.1:{port}", proc

def call_vllm_worker(messages, model_name, worker_addrs, timeout:int=300, conv_system_msg=None, **generate_kwargs) -> str:
    global worker_initiated
    global chat_tokenizers
    if "max_new_tokens" in generate_kwargs:
        if "max_tokens" not in generate_kwargs:
            generate_kwargs["max_tokens"] = generate_kwargs["max_new_tokens"]
        del generate_kwargs["max_new_tokens"]
    try:
        if model_name not in chat_tokenizers:
            chat_tokenizers[model_name] = ChatTokenizer(model_name)
        chat_tokenizer = chat_tokenizers[model_name]
        prompt = chat_tokenizer(chat_messages)
    except Exception as e:
        pass
    
    chat_messages = []
    if conv_system_msg:
        chat_messages.append({"role": "system", "content": conv_system_msg})
    for i, message in enumerate(messages):
        chat_messages.append({"role": "user" if i % 2 == 0 else "assistant", "content": message})


    worker_addr = random.choice(worker_addrs)
    
    client = openai.OpenAI(
        base_url=f"{worker_addr}/v1",
        # api_key="vllm-engine-token",
    )

    args_names, kwargs_names = get_function_arg_names(client.chat.completions.create)
    extra_body_params = {}
    for key in list(generate_kwargs.keys()):
        if key not in args_names + kwargs_names:
            extra_body_params[key] = generate_kwargs[key]
            del generate_kwargs[key]
    generate_kwargs["extra_body"] = extra_body_params
    
    stream = generate_kwargs.get("stream", False)
    if stream:
        generate_kwargs.pop("n", None)
    
    print(generate_kwargs)
    @with_timeout(timeout)
    def get_response():
        while True:
            try:
                completion = client.chat.completions.create(
                    model=model_name,
                    messages=chat_messages,
                    **generate_kwargs,
                )
                break
            except openai.APIConnectionError as e:
                if not worker_initiated:
                    time.sleep(5)
                    continue
                print(f"API connection error: {e}")
                time.sleep(5)
                continue
            except openai.BadRequestError as e:
                error_response = e.response.json()
                if error_response['code'] == 400:
                    pattern = r"This model's maximum context length is (\d+) tokens. However, you requested (\d+) tokens \((\d+) in the messages, (\d+) in the completion\). Please reduce the length of the messages or completion."
                    res = re.match(pattern, error_response['message'])
                    if res:
                        max_context_length = int(res.group(1))
                        num_tokens_requested = int(res.group(2))
                        num_tokens_in_messages = int(res.group(3))
                        num_tokens_in_completion = int(res.group(4))
                        
                        new_max_tokens = num_tokens_in_completion - (num_tokens_requested - max_context_length)
                        if new_max_tokens <= 0:
                            raise e
                        print(f"Reducing max_tokens to {new_max_tokens}, and retrying")
                        generate_kwargs["max_tokens"] = new_max_tokens
                        continue
                    else:
                        raise e

        if not stream:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices]
            else:
                return completion.choices[0].message.content
        else:
            def generate_stream():
                for chunk in completion:
                    if chunk.choices[0].delta.content is not None:
                        yield chunk.choices[0].delta.content
            return generate_stream()
    
    return get_response()
    
def call_vllm_worker_completion(prompt:str, model_name, worker_addrs, timeout:int=300, **generate_kwargs) -> str:
    global worker_initiated
    
    worker_addr = random.choice(worker_addrs)
    
    client = openai.OpenAI(
        base_url=f"{worker_addr}/v1",
        api_key="vllm-engine-token",
    )
    
    stream = generate_kwargs.get("stream", False)
    if stream:
        generate_kwargs.pop("n", None)
        
    args_names, kwargs_names = get_function_arg_names(client.completions.create)
    extra_body_params = {}
    for key in list(generate_kwargs.keys()):
        if key not in args_names + kwargs_names:
            extra_body_params[key] = generate_kwargs[key]
            del generate_kwargs[key]
    # generate_kwargs["extra_body"] = extra_body_params
    
    @with_timeout(timeout)
    def get_response():
        while True:
            try:
                completion = client.completions.create(
                    model=model_name,
                    prompt=prompt,
                    **generate_kwargs,
                )
                break
            except openai.APIConnectionError as e:
                if not worker_initiated:
                    time.sleep(5)
                    continue
                print(f"API connection error: {e}")
                time.sleep(5)
                continue
        if not stream:
            if len(completion.choices) > 1:
                return [c.text for c in completion.choices]
            else:
                return completion.choices[0].text
        else:
            def generate_stream():
                for chunk in completion:
                    if chunk.choices[0].text is not None:
                        yield chunk.choices[0].text
            return generate_stream()
    return get_response()


from openai import OpenAI
from typing import List

# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_openai(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to openai format
    new_messages = []
    if conv_system_msg:
        new_messages.append({"role": "system", "content": conv_system_msg})
    for i, message in enumerate(messages):
        new_messages.append({"role": "user" if i % 2 == 0 else "assistant", "content": message})
    # initialize openai client
    client = OpenAI()
    # call openai
    completion = client.chat.completions.create(
        model=model_name,
        messages=new_messages,
        timeout=timeout,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    if not stream:
        if len(completion.choices) > 1:
            return [c.message.content for c in completion.choices]
        else:
            return completion.choices[0].message.content
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        return generate_stream()

def call_worker_openai_completion(prompt:str, model_name, timeout:int=60, **generate_kwargs) -> str:
    # initialize openai client
    client = OpenAI()
    # call openai
    completion = client.completions.create(
        model=model_name,
        prompt=prompt,
        timeout=timeout,
        **generate_kwargs,
    )
    stream = generate_kwargs.get("stream", False)
    if not stream:
        if len(completion.choices) > 1:
            return [c.text for c in completion.choices]
        else:
            return completion.choices[0].text
    else:
        def generate_stream():
            for chunk in completion:
                if chunk.choices[0].text is not None:
                    yield chunk.choices[0].text
        return generate_stream()

if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_openai(["Hello"], "gpt-3.5-turbo"))
    ic(call_worker_openai_completion("Hello", "gpt-3.5-turbo-instruct"))


import json
import os
import shutil
import atexit
import sqlite3
import mmap
import threading
import struct
from .cache_utils import get_cache_file
from pathlib import Path
from typing import Union, List, Dict
from cachetools import LRUCache
from tqdm import tqdm

BLOCK_SIZE = 3072 * 1024  # 3MB
MAX_CACHE_SIZE = 10000  # Example: 10k items
MAX_MEMORY_BLOCKS = 32 # Example: 32 blocks
# Global cache dictionary using MultiLevelCache
cache_dict = {}

class BlockCache:
    def __init__(self, max_size=MAX_MEMORY_BLOCKS):
        self.cache = LRUCache(maxsize=max_size)
        self.lock = threading.Lock()

    def get(self, block_id):
        with self.lock:
            return self.cache.get(block_id)

    def set(self, block_id, data):
        with self.lock:
            self.cache[block_id] = data

class EfficientDiskCache:
    def __init__(self, cache_dir, model_name, block_size=BLOCK_SIZE, max_memory_blocks=MAX_MEMORY_BLOCKS):
        self.cache_dir = Path(cache_dir) / f"{model_name}_disk_cache"
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.model_name = model_name
        self.block_size = block_size
        self.index_db = self.cache_dir / "index.db"
        self.block_cache = BlockCache(max_size=max_memory_blocks)
        self.db_lock = threading.Lock()
        self.init_index_db()

    def __del__(self):
        self.cleanup()

    def cleanup(self):
        try:
            if self.cache_dir.exists():
                shutil.rmtree(self.cache_dir)
            print(f"Cleaned up cache directory: {self.cache_dir}")
        except Exception as e:
            print(f"Error during cleanup: {e}")

    def init_index_db(self):
        with self.db_lock:
            with sqlite3.connect(self.index_db) as conn:
                cursor = conn.cursor()
                cursor.execute('''
                    CREATE TABLE IF NOT EXISTS cache_index
                    (key TEXT PRIMARY KEY, block_id INTEGER, offset INTEGER, length INTEGER)
                ''')
                conn.commit()

    def get_block_file(self, block_id):
        return self.cache_dir / f"block_{block_id}.bin"

    def get(self, key):
        with self.db_lock:
            with sqlite3.connect(self.index_db) as conn:
                cursor = conn.cursor()
                cursor.execute("SELECT block_id, offset, length FROM cache_index WHERE key = ?", (key,))
                result = cursor.fetchone()
        
        if result:
            block_id, offset, length = result
            block_data = self.block_cache.get(block_id)
            if block_data is None:
                block_file = self.get_block_file(block_id)
                if block_file.exists():
                    try:
                        with open(block_file, 'rb') as f:
                            mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
                            block_data = mm.read()
                        self.block_cache.set(block_id, block_data)
                    except Exception as e:
                        print(f"Error reading block file: {e}")
                        return None
                else:
                    return None
            
            try:
                data = block_data[offset:offset+length]
                value_length = struct.unpack('!I', data[:4])[0]
                json_data = data[4:4+value_length].decode('utf-8')
                return json.loads(json_data)
            except (struct.error, json.JSONDecodeError, UnicodeDecodeError) as e:
                print(f"Error decoding data: {e}")
                return None
        return None

    def set(self, key, value):
        json_data = json.dumps(value).encode('utf-8')
        data_length = len(json_data)
        full_data = struct.pack('!I', data_length) + json_data

        with self.db_lock:
            with sqlite3.connect(self.index_db) as conn:
                cursor = conn.cursor()
                cursor.execute("SELECT MAX(block_id) FROM cache_index")
                result = cursor.fetchone()
                current_block_id = result[0] if result[0] is not None else 0
                
                block_file = self.get_block_file(current_block_id)
                if block_file.exists():
                    file_size = block_file.stat().st_size
                    if file_size + len(full_data) > self.block_size:
                        current_block_id += 1
                        block_file = self.get_block_file(current_block_id)
                        file_size = 0
                else:
                    file_size = 0

                with open(block_file, 'ab') as f:
                    f.write(full_data)
                    
                cursor.execute('''
                    INSERT OR REPLACE INTO cache_index (key, block_id, offset, length)
                    VALUES (?, ?, ?, ?)
                ''', (key, current_block_id, file_size, len(full_data)))
                conn.commit()

        block_data = self.block_cache.get(current_block_id)
        if block_data is not None:
            self.block_cache.set(current_block_id, block_data + full_data)

    def bulk_insert(self, data: Dict[str, Dict]):
        with self.db_lock:
            with sqlite3.connect(self.index_db) as conn:
                cursor = conn.cursor()
                cursor.execute("SELECT MAX(block_id) FROM cache_index")
                result = cursor.fetchone()
                current_block_id = result[0] if result[0] is not None else 0
                
                block_file = self.get_block_file(current_block_id)
                if block_file.exists():
                    file_size = block_file.stat().st_size
                else:
                    file_size = 0

                index_data = []
                current_block_data = b""
                
                for key, value in data.items():
                    json_data = json.dumps(value).encode('utf-8')
                    data_length = len(json_data)
                    full_data = struct.pack('!I', data_length) + json_data
                    
                    if file_size + len(full_data) > self.block_size:
                        with open(block_file, 'ab') as f:
                            f.write(current_block_data)
                        self.block_cache.set(current_block_id, current_block_data)
                        
                        current_block_id += 1
                        block_file = self.get_block_file(current_block_id)
                        file_size = 0
                        current_block_data = b""
                    
                    index_data.append((key, current_block_id, file_size, len(full_data)))
                    current_block_data += full_data
                    file_size += len(full_data)

                if current_block_data:
                    with open(block_file, 'ab') as f:
                        f.write(current_block_data)
                    self.block_cache.set(current_block_id, current_block_data)

                cursor.executemany('''
                    INSERT OR REPLACE INTO cache_index (key, block_id, offset, length)
                    VALUES (?, ?, ?, ?)
                ''', index_data)
                conn.commit()
                            
class MultiLevelCache:
    def __init__(self, model_name, cache_dir, memory_size=MAX_CACHE_SIZE):
        self.memory_cache = LRUCache(maxsize=memory_size)
        self.disk_cache = EfficientDiskCache(cache_dir, model_name)

    def get(self, key):
        value = self.memory_cache.get(key)
        if value is not None:
            return value

        value = self.disk_cache.get(key)
        if value is not None:
            self.memory_cache[key] = value
            return value

        return None
    
    def __getitem__(self, key):
        return self.get(key)

    def set(self, key, value):
        self.memory_cache[key] = value
        self.disk_cache.set(key, value)
        
    def __setitem__(self, key, value):
        self.set(key, value)

    def bulk_insert(self, data: Dict[str, Dict]):
        self.disk_cache.bulk_insert(data)
        for key, value in tqdm(data.items(), desc="Bulk inserting into Memory Cache"):
            self.memory_cache[key] = value


def load_cache(model_name, cache_dir=None):
    global cache_dict
    if model_name not in cache_dict:
        if cache_dir is None:
            cache_dir = Path(os.path.expanduser(f"~/llm_engines/generation_cache"))
        else:
            cache_dir = Path(cache_dir)
        cache_file = get_cache_file(model_name, cache_dir)
        cache_dict[model_name] = MultiLevelCache(model_name, cache_dir)
        
        if cache_file.exists():
            print("Cache file exists at:", cache_file.absolute())
            initial_data = {}
            with open(cache_file, 'r') as f:
                for line in tqdm(f, desc="Loading cache for model: " + model_name):
                    data = json.loads(line)
                    key = list(data.keys())[0]  
                    initial_data[key] = data[key]
            if initial_data:
                cache_dict[model_name].bulk_insert(initial_data)
    return cache_dict[model_name]

# Cleanup function to be called at exit
def cleanup_all_caches():
    global cache_dict
    print("Cleaning up all caches...")
    for model_name, cache in cache_dict.items():
        cache.disk_cache.cleanup()
    cache_dict.clear()

# Register the cleanup function to be called at exit
atexit.register(cleanup_all_caches)


import os
import time
import torch
import random
import openai
import importlib.util
from pathlib import Path
from typing import List
from sglang import function, system, user, assistant, gen
from sglang.srt.utils import allocate_init_ports
from .utils import SubprocessMonitor, ChatTokenizer, with_timeout, get_function_arg_names
worker_initiated = False
sglang_workers = {}
def launch_sglang_worker(
    model_name: str,
    num_gpus: int=None,
    gpu_ids: List[int]=None,
    dtype: str="auto",
    quantization: str=None,
    port: int=34200,
    host: str="127.0.0.1",
    additional_args: List[str]=[]
) -> str:
    """
    Launch a model worker and return the address
    Args:
        model_name: the model name to launch
    Returns:
        the address of the launched model
    """
    # python -m sglang.launch_server --model-path meta-llama/Meta-Llama-3-8B-Instruct --port 30000
    ### For debug
    # port, additonal_ports = allocate_init_ports(port)
    # print(f"Launching SGLang model {model_name} on port {port}")
    # print(f"Additional ports: {additonal_ports}")
    ### For debug
    worker_addr = f"http://{host}:{port}"
    log_file = Path(os.path.abspath(__file__)).parent / "logs" / f"{model_name}.log"
    log_file.parent.mkdir(parents=True, exist_ok=True)
    if gpu_ids:
        num_gpus = len(gpu_ids)
    else:
        if not num_gpus:
            num_gpus = torch.cuda.device_count()
            print(f"Warning: num_gpus or gpu_ids not provided, using {num_gpus} GPUs")
        gpu_ids = list(range(num_gpus))
    env = os.environ.copy()
    # Set the CUDA_VISIBLE_DEVICES environment variable
    env["CUDA_VISIBLE_DEVICES"] = ",".join([str(gpu_id) for gpu_id in gpu_ids])
    
    # check flashinfer
    flashinfer = importlib.util.find_spec("flashinfer")
    if flashinfer is None:
        print("flashinfer not found, please first install flashinfer for sglang")
        print("Simple Command: pip install flashinfer -i https://flashinfer.ai/whl/cu121/torch2.4/")
        print("Please refer to https://docs.flashinfer.ai/installation.html for detailed installation instructions")
        exit(1) 
    else:
        print("flashinfer found, enable flashinfer for sglang")
        flashinfer_args = []
    if quantization:
        available_quantizations = "awq,fp8,gptq,marlin,gptq_marlin,awq_marlin,squeezellm,bitsandbytes"
        available_quantizations = available_quantizations.split(",")
        if quantization not in available_quantizations:
            raise ValueError(f"Quantization {quantization} not supported, available quantizations: {available_quantizations}")
        flashinfer_args = ["--quantization", quantization]
    # additonal_ports = [port+i for i in range(1, 9)]
    proc = SubprocessMonitor([
        "python3", "-m", "sglang.launch_server",
        "--model-path", model_name,
        "--host", host,
        "--port", str(port),
        "--dtype", dtype,
        # "--api-key", "sglang",
        "--log-level", "warning",
        "--tp-size",  str(num_gpus) if num_gpus is not None else "1",
        # "--additional-ports"] + [str(port) for port in additonal_ports
    ] + flashinfer_args + additional_args ,env=env)
    print(f"Launching SGLang model {model_name} with CUDA_VISIBLE_DEVICES={env['CUDA_VISIBLE_DEVICES']}")
    sglang_workers[worker_addr] = proc
    return worker_addr, proc

@function
def multi_turn_question(s, messages, system_message=None):
    if system_message:
        s += system(system_message)
    for i, message in enumerate(messages):
        if i % 2 == 0:
            s += user(message)
        else:
            s += assistant(message)
    s += assistant(gen("answer"))
    
@function
def question(s, prompt):
    s += prompt
    s += gen("answer")

chat_tokenizers = {}
def call_sglang_worker(messages, model_name, worker_addrs, timeout:int=300, conv_system_msg=None, **generate_kwargs) -> str:
    global worker_initiated
    global chat_tokenizers
    
    if model_name not in chat_tokenizers:
        chat_tokenizers[model_name] = ChatTokenizer(model_name)
    chat_tokenizer = chat_tokenizers[model_name]
    
    chat_messages = []
    if conv_system_msg:
        chat_messages.append({"role": "system", "content": conv_system_msg})
    for i, message in enumerate(messages):
        chat_messages.append({"role": "user" if i % 2 == 0 else "assistant", "content": message})

    prompt = chat_tokenizer(chat_messages)

    worker_addr = random.choice(worker_addrs)
    
    client = openai.OpenAI(
        base_url=f"{worker_addr}/v1",
        api_key="sglang-engine-token",
    )
    
    generate_kwargs['max_tokens'] = generate_kwargs.get('max_tokens', chat_tokenizer.max_length) # for sglang, max_tokens is required and must > 0
    args_names, kwargs_names = get_function_arg_names(client.chat.completions.create)
    extra_body_params = {}
    for key in list(generate_kwargs.keys()):
        if key not in args_names + kwargs_names:
            extra_body_params[key] = generate_kwargs[key]
            del generate_kwargs[key]
    generate_kwargs["extra_body"] = extra_body_params
    
    stream = generate_kwargs.get("stream", False)
    if stream:
        generate_kwargs.pop("n", None)
    @with_timeout(timeout)
    def get_response():
        while True:
            try:
                print()
                completion = client.chat.completions.create(
                    model=model_name,
                    messages=chat_messages,
                    **generate_kwargs,
                )
                break
            except openai.APIConnectionError as e:
                if not worker_initiated:
                    time.sleep(5)
                    continue
                print(f"API connection error: {e}")
                time.sleep(5)
                continue
        if not stream:
            if len(completion.choices) > 1:
                return [c.message.content for c in completion.choices]
            else:
                return completion.choices[0].message.content
        else:
            def generate_stream():
                for chunk in completion:
                    if chunk.choices[0].delta.content is not None:
                        yield chunk.choices[0].delta.content
            return generate_stream()
    return get_response()

def call_sglang_worker_completion(prompt:str, model_name, worker_addrs, timeout:int=300, **generate_kwargs) -> str:
    global worker_initiated
    global chat_tokenizers
    if model_name not in chat_tokenizers:
        chat_tokenizers[model_name] = ChatTokenizer(model_name)
    chat_tokenizer = chat_tokenizers[model_name]
    
    if "max_new_tokens" in generate_kwargs:
        if "max_tokens" not in generate_kwargs:
            generate_kwargs["max_tokens"] = generate_kwargs["max_new_tokens"]
        del generate_kwargs["max_new_tokens"]
        
    worker_addr = random.choice(worker_addrs)
    
    client = openai.OpenAI(
        base_url=f"{worker_addr}/v1",
        api_key="sglang-engine-token",
    )
    
    generate_kwargs['max_tokens'] = generate_kwargs.get('max_tokens', chat_tokenizer.max_length) # for sglang, max_tokens is required and must > 0
    args_names, kwargs_names = get_function_arg_names(client.completions.create)
    extra_body_params = {}
    for key in list(generate_kwargs.keys()):
        if key not in args_names + kwargs_names:
            extra_body_params[key] = generate_kwargs[key]
            del generate_kwargs[key]
    generate_kwargs["extra_body"] = extra_body_params
    
    stream = generate_kwargs.get("stream", False)
    if stream:
        generate_kwargs.pop("n", None)
    @with_timeout(timeout)
    def get_response():
        while True:
            try:
                completion = client.completions.create(
                    model=model_name,
                    prompt=prompt,
                    **generate_kwargs,
                )
                break
            except openai.APIConnectionError as e:
                if not worker_initiated:
                    time.sleep(5)
                    continue
                print(f"API connection error: {e}")
                time.sleep(5)
                continue
        
        # return completion.choices[0].text
        if not stream:
            if len(completion.choices) > 1:
                return [c.text for c in completion.choices]
            else:
                return completion.choices[0].text
        else:
            def generate_stream():
                for chunk in completion:
                    if chunk.choices[0].text is not None:
                        yield chunk.choices[0].text
            return generate_stream()
    return get_response()


import anthropic
import os
from typing import List
from anthropic import NOT_GIVEN
from .utils import with_timeout

# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_claude(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to mistral format
    client = anthropic.Anthropic(api_key=os.environ.get("ANTHROPIC_API_KEY"))
    new_messages = []
    for i, message in enumerate(messages):
        new_messages.append({"role": "user" if i % 2 == 0 else "assistant", "content": message})
             
    generate_kwargs.pop("n", None) # claude does not have n
    if not generate_kwargs.get("max_tokens", None):
        generate_kwargs["max_tokens"] = 1024
    stream = generate_kwargs.pop("stream", False)
    @with_timeout(timeout)
    def get_response():
        completion = client.messages.create(
            model=model_name,
            messages=new_messages,
            system=conv_system_msg if conv_system_msg else NOT_GIVEN,
            timeout=timeout,
            **generate_kwargs,
        )
        if len(completion.content) > 1:
            return [c.text for c in completion.content]
        else:
            return completion.content[0].text
        
    @with_timeout(timeout)
    def stream_response():
        with client.messages.stream(
            model=model_name,
            messages=new_messages,
            system=conv_system_msg if conv_system_msg else NOT_GIVEN,
            timeout=timeout,
            **generate_kwargs,
        ) as stream:
            for text in stream.text_stream:
                yield text
                
    if not stream:
        return get_response()
    else:
        return stream_response()
    
if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_claude(["Hello", "Hi, I am claude", "What did I ask in the last response?"], "claude-3-opus-20240229"))


import os
from typing import List
import google.ai.generativelanguage as glm
import google.generativeai as genai
from .utils import with_timeout
genai.configure(api_key=os.environ.get("GOOGLE_API_KEY"))

safety_settings = [
    {
        "category": "HARM_CATEGORY_DANGEROUS",
        "threshold": "BLOCK_NONE",
    },
    {
        "category": "HARM_CATEGORY_HARASSMENT",
        "threshold": "BLOCK_NONE",
    },
    {
        "category": "HARM_CATEGORY_HATE_SPEECH",
        "threshold": "BLOCK_NONE",
    },
    {
        "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
        "threshold": "BLOCK_NONE",
    },
    {
        "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
        "threshold": "BLOCK_NONE",
    },
]
# no image, multi-turn, do not use openai_generate, but can refer to it
def call_worker_gemini(messages:List[str], model_name, timeout:int=60, conv_system_msg=None, **generate_kwargs) -> str:
    # change messages to gemini format
    model = genai.GenerativeModel(model_name, system_instruction=conv_system_msg)
    
    new_messages = []
    for i, message in enumerate(messages):
        new_messages.append({"role": "user" if i % 2 == 0 else "model", "parts": [glm.Part(text=message)]})
    
    stream = generate_kwargs.pop("stream", False)
    generation_config = genai.GenerationConfig(
        candidate_count=generate_kwargs.get("num_return_sequences", None),
        stop_sequences=generate_kwargs.get("stop", None),
        max_output_tokens=generate_kwargs.get("max_tokens", None),
        temperature=generate_kwargs.get("temperature", None),
        top_p=generate_kwargs.get("top_p", None),
        top_k=generate_kwargs.get("top_k", None),
        response_mime_type=generate_kwargs.get("response_mime_type", None),
        response_schema=generate_kwargs.get("response_schema", None),
    )
    request_options = genai.types.RequestOptions(
        timeout=timeout,
    )
    @with_timeout(timeout)
    def generate_content():
        return model.generate_content(new_messages, safety_settings=safety_settings, generation_config=generation_config, request_options=request_options, stream=stream)
    response = generate_content()
    if not stream:
        return response.text
    else:
        def generate_stream():
            for chunk in response:
                yield chunk.text
        return generate_stream()

if __name__ == "__main__":
    from icecream import ic
    ic(call_worker_gemini(["Hello", "Hi, I am gemini", "What did I ask in the last response?"], "gemini-1.5-flash"))



from setuptools import setup, find_packages
# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='llm-engines',
    version='0.0.11',
    description='A unified inference engine for large language models (LLMs) including open-source models (VLLM, SGLang, Together) and commercial models (OpenAI, Mistral, Claude).',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Dongfu Jiang',
    author_email='dongfu.jiang@uwaterloo.ca',
    packages=find_packages(),
    url='https://github.com/jdf-progLLM-Engines',
    entry_points={"console_scripts": ["llm-engines=llm_engines.cli:main"]},
    install_requires=[
        "fire",
        "openai",
        "google-generativeai",
        "accelerate",
        "transformers",
        "torch",
        "Pillow",
        "torch",
        "tqdm",
        "numpy",
        "requests",
        "sentencepiece",
        "vllm>=0.5.5",
        "together",
        "icecream",
        "prettytable",
        "sglang[all]",
        "mistralai",
        "anthropic",
        "fastapi<=0.112.2"
    ],
    extras_require={
        "flash-attn": {
            "flash-attn"
        }
    }
)

"""
rm -rf dist build llm_engines.egg-info
python setup.py sdist bdist_wheel
twine upload dist/*
"""